#include "ECC.hpp"
#include "BigInt.hpp"
#include "Sha2.hpp"
#include "Random.hpp"
using namespace big;

/*
	HisilWongCarterDawson paper "Twisted Edwards Curves Revisited" (Asiacrypt 2008)

	Twisted Edwards E(p) Curve: a * x^2 + y^2 = 1 + d * x^2 * y^2, d = (given), a = -1, p = 2^256 - c

	Edwards coordinates: (X : Y : Z)
	Extended Edwards coordinates: (X : Y : T : Z), T = XY
	Edwards to Extended Edwards: (X : Y : Z) -> (XZ : YZ : XY : ZZ)
	Extended Edwards to Edwards: (X : Y : T : Z) -> (X : Y : Z)

	-(X : Y : T : Z) = (-X : Y : -T : Z)

	Identity element: (0 : 1 : 0 : 1)

	When Z = 1, a multiplication can be omitted

	Doubling, doubling -> E = 2E
	Doubling, add -> Ee = 2E, E = Ee + Ee
*/

// Returns x = X/Z
void EEdGetAffineX(
	const EdPoint *p1,
	u32 *x)
{
	u32 p[EDWARD_LIMBS*2], m[EDWARD_LIMBS], A[EDWARD_LIMBS];

	// m = modulus
	Set32(m, EDWARD_LIMBS, 0);
	Subtract32(m, EDWARD_LIMBS, EDWARD_C);

	// A = 1/Z (mod m)
	InvMod(p1->Z, EDWARD_LIMBS, m, EDWARD_LIMBS, A);

	// x = X * A
	Multiply(EDWARD_LIMBS, p, p1->X, A);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, x);
}

// Returns (x,y) = (X/Z,Y/Z)
void EEdGetAffineXY(
	const EdPoint *p1,
	u32 *x,
	u32 *y)
{
	u32 p[EDWARD_LIMBS*2], m[EDWARD_LIMBS], A[EDWARD_LIMBS];

	// m = modulus
	Set32(m, EDWARD_LIMBS, 0);
	Subtract32(m, EDWARD_LIMBS, EDWARD_C);

	// A = 1/Z (mod m)
	InvMod(p1->Z, EDWARD_LIMBS, m, EDWARD_LIMBS, A);

	// x = X * A
	Multiply(EDWARD_LIMBS, p, p1->X, A);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, x);

	// y = Y * A
	Multiply(EDWARD_LIMBS, p, p1->Y, A);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, y);
}

// Solve for Y given the X point on a curve
void EEdSolveY(
	const u32 *x,
	u32 *y)
{
	u32 p[EDWARD_LIMBS*2], m[EDWARD_LIMBS];
	u32 A[EDWARD_LIMBS], B[EDWARD_LIMBS], C[EDWARD_LIMBS], D[EDWARD_LIMBS], E[EDWARD_LIMBS];

	// m = modulus
	Set32(m, EDWARD_LIMBS, 0);
	Subtract32(m, EDWARD_LIMBS, EDWARD_C);

	// A = x^2
	Square(EDWARD_LIMBS, p, x);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, A);

	// B = A + 1
	Set(B, EDWARD_LIMBS, A);
	Add32(B, EDWARD_LIMBS, 1);

	// C = d * A
	p[EDWARD_LIMBS] = Multiply32(EDWARD_LIMBS, p, A, EDWARD_D);
	SpecialModulus(p, EDWARD_LIMBS+1, EDWARD_C, EDWARD_LIMBS, C);

	// C = -C
	Negate(EDWARD_LIMBS, C, C);
	Subtract32(C, EDWARD_LIMBS, EDWARD_C);

	// C = C + 1
	Add32(C, EDWARD_LIMBS, 1);

	// D = 1/C (mod m)
	InvMod(C, EDWARD_LIMBS, m, EDWARD_LIMBS, D);

	// E = B * D
	Multiply(EDWARD_LIMBS, p, B, D);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, E);

	// return square root
	SpecialSquareRoot(EDWARD_LIMBS, E, EDWARD_C, y);
}

// Verify that the point (x,y) exists on the given curve
bool EEdIsValidXY(
	const u32 *x,
	const u32 *y)
{
	// 0 = 1 + d*x^2*y^2 + x^2 - y^2
	// 0 = 1 + dAB + A - B
	// 0 = 1 + Z + A - B
	u32 p[EDWARD_LIMBS*2+1];
	u32 A[EDWARD_LIMBS], B[EDWARD_LIMBS], Z[EDWARD_LIMBS];

	// A = x^2, B = y^2
	Square(EDWARD_LIMBS, p, x);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, A);
	Square(EDWARD_LIMBS, p, y);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, B);

	// Z = dAB
	Multiply(EDWARD_LIMBS, p, A, B);
	p[EDWARD_LIMBS*2] = Multiply32(EDWARD_LIMBS*2, p, p, EDWARD_D);
	SpecialModulus(p, EDWARD_LIMBS*2+1, EDWARD_C, EDWARD_LIMBS, Z);

	// Z += A
	if (Add(Z, EDWARD_LIMBS, A, EDWARD_LIMBS))
		Add32(Z, EDWARD_LIMBS, EDWARD_C);

	// Z += 1
	if (Add32(Z, EDWARD_LIMBS, 1))
		Add32(Z, EDWARD_LIMBS, EDWARD_C);

	// Z -= B
	if (Subtract(Z, EDWARD_LIMBS, B, EDWARD_LIMBS))
		Subtract32(Z, EDWARD_LIMBS, EDWARD_C);

	// if Z = 0, success
	if (!LimbDegree(Z, EDWARD_LIMBS))
		return true;

	// if Z = modulus, success
	Add32(Z, EDWARD_LIMBS, EDWARD_C);
	if (!LimbDegree(Z, EDWARD_LIMBS))
		return true;

	// otherwise it is not on the curve
	return false;
}

// Extended Twisted Edwards Negation Formula in 2a
void EEdNegate(
   const EdPoint *p1,
   EdPoint *p3)
{
	// -(X : Y : T : Z) = (-X : Y : -T : Z)

	// -X
	Negate(EDWARD_LIMBS, p3->X, p1->X);
	Subtract32(p3->X, EDWARD_LIMBS, EDWARD_C);

	// Y
	Set(p3->Y, EDWARD_LIMBS, p1->Y);

	// -T
	Negate(EDWARD_LIMBS, p3->T, p1->T);
	Subtract32(p3->T, EDWARD_LIMBS, EDWARD_C);

	// Z
	Set(p3->Z, EDWARD_LIMBS, p1->Z);
}

// Extended Twisted Edwards Unified Addition Formula (works when both inputs are the same) in 8M 1d 8a
void EEdAdd(
   const EdPoint *p1,
   const EdPoint *p2,
   EdPoint *p3)
{
	u32 A[EDWARD_LIMBS], B[EDWARD_LIMBS], C[EDWARD_LIMBS], D[EDWARD_LIMBS];
	u32 E[EDWARD_LIMBS], F[EDWARD_LIMBS], G[EDWARD_LIMBS], H[EDWARD_LIMBS];
	u32 p[EDWARD_LIMBS*2+1];

	// A = (Y1 - X1) * (Y2 - X2)
	if (Subtract(C, p1->Y, EDWARD_LIMBS, p1->X, EDWARD_LIMBS))
		Subtract32(C, EDWARD_LIMBS, EDWARD_C);
	if (Subtract(D, p2->Y, EDWARD_LIMBS, p2->X, EDWARD_LIMBS))
		Subtract32(D, EDWARD_LIMBS, EDWARD_C);
	Multiply(EDWARD_LIMBS, p, C, D);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, A);

	// B = (Y1 + X1) * (Y2 + X2)
	if (Add(C, p1->Y, EDWARD_LIMBS, p1->X, EDWARD_LIMBS))
		Add32(C, EDWARD_LIMBS, EDWARD_C);
	if (Add(D, p2->Y, EDWARD_LIMBS, p2->X, EDWARD_LIMBS))
		Add32(D, EDWARD_LIMBS, EDWARD_C);
	Multiply(EDWARD_LIMBS, p, C, D);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, B);

	// C = 2 * d' * T1 * T2 (can remove multiplication by d' if inputs are known to be different)
	Multiply(EDWARD_LIMBS, p, p1->T, p2->T);
	p[EDWARD_LIMBS*2] = Multiply32(EDWARD_LIMBS*2, p, p, EDWARD_D*2);
	SpecialModulus(p, EDWARD_LIMBS*2+1, EDWARD_C, EDWARD_LIMBS, C);

	// D = 2 * Z1 * Z2
	Multiply(EDWARD_LIMBS, p, p1->Z, p2->Z);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, D);
	if (ShiftLeft(EDWARD_LIMBS, D, D, 1))
		Add32(D, EDWARD_LIMBS, EDWARD_C);

	// E = B - A, F = D - C, G = D + C, H = B + A
	if (Subtract(E, B, EDWARD_LIMBS, A, EDWARD_LIMBS))
		Subtract32(E, EDWARD_LIMBS, EDWARD_C);
	if (Subtract(F, D, EDWARD_LIMBS, C, EDWARD_LIMBS))
		Subtract32(F, EDWARD_LIMBS, EDWARD_C);
	if (Add(G, D, EDWARD_LIMBS, C, EDWARD_LIMBS))
		Add32(G, EDWARD_LIMBS, EDWARD_C);
	if (Add(H, B, EDWARD_LIMBS, A, EDWARD_LIMBS))
		Add32(H, EDWARD_LIMBS, EDWARD_C);

	// X3 = E * F
	Multiply(EDWARD_LIMBS, p, E, F);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->X);

	// Y3 = G * H
	Multiply(EDWARD_LIMBS, p, G, H);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->Y);

	// T3 = E * H
	Multiply(EDWARD_LIMBS, p, E, H);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->T);

	// Z3 = F * G
	Multiply(EDWARD_LIMBS, p, F, G);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->Z);
}

// Extended Twisted Edwards Unified Addition Formula (works when both inputs are the same) in 7M 1d 8a
void EEdAddNoT(
   const EdPoint *p1,
   const EdPoint *p2,
   EdPoint *p3)
{
	u32 A[EDWARD_LIMBS], B[EDWARD_LIMBS], C[EDWARD_LIMBS], D[EDWARD_LIMBS];
	u32 E[EDWARD_LIMBS], F[EDWARD_LIMBS], G[EDWARD_LIMBS], H[EDWARD_LIMBS];
	u32 p[EDWARD_LIMBS*2+1];

	// A = (Y1 - X1) * (Y2 - X2)
	if (Subtract(C, p1->Y, EDWARD_LIMBS, p1->X, EDWARD_LIMBS))
		Subtract32(C, EDWARD_LIMBS, EDWARD_C);
	if (Subtract(D, p2->Y, EDWARD_LIMBS, p2->X, EDWARD_LIMBS))
		Subtract32(D, EDWARD_LIMBS, EDWARD_C);
	Multiply(EDWARD_LIMBS, p, C, D);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, A);

	// B = (Y1 + X1) * (Y2 + X2)
	if (Add(C, p1->Y, EDWARD_LIMBS, p1->X, EDWARD_LIMBS))
		Add32(C, EDWARD_LIMBS, EDWARD_C);
	if (Add(D, p2->Y, EDWARD_LIMBS, p2->X, EDWARD_LIMBS))
		Add32(D, EDWARD_LIMBS, EDWARD_C);
	Multiply(EDWARD_LIMBS, p, C, D);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, B);

	// C = 2 * d' * T1 * T2 (can remove multiplication by d' if inputs are known to be different)
	Multiply(EDWARD_LIMBS, p, p1->T, p2->T);
	p[EDWARD_LIMBS*2] = Multiply32(EDWARD_LIMBS*2, p, p, EDWARD_D*2);
	SpecialModulus(p, EDWARD_LIMBS*2+1, EDWARD_C, EDWARD_LIMBS, C);

	// D = 2 * Z1 * Z2
	Multiply(EDWARD_LIMBS, p, p1->Z, p2->Z);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, D);
	if (ShiftLeft(EDWARD_LIMBS, D, D, 1))
		Add32(D, EDWARD_LIMBS, EDWARD_C);

	// E = B - A, F = D - C, G = D + C, H = B + A
	if (Subtract(E, B, EDWARD_LIMBS, A, EDWARD_LIMBS))
		Subtract32(E, EDWARD_LIMBS, EDWARD_C);
	if (Subtract(F, D, EDWARD_LIMBS, C, EDWARD_LIMBS))
		Subtract32(F, EDWARD_LIMBS, EDWARD_C);
	if (Add(G, D, EDWARD_LIMBS, C, EDWARD_LIMBS))
		Add32(G, EDWARD_LIMBS, EDWARD_C);
	if (Add(H, B, EDWARD_LIMBS, A, EDWARD_LIMBS))
		Add32(H, EDWARD_LIMBS, EDWARD_C);

	// X3 = E * F
	Multiply(EDWARD_LIMBS, p, E, F);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->X);

	// Y3 = G * H
	Multiply(EDWARD_LIMBS, p, G, H);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->Y);
/*
	// T3 = E * H
	Multiply(EDWARD_LIMBS, p, E, H);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->T);
*/
	// Z3 = F * G
	Multiply(EDWARD_LIMBS, p, F, G);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->Z);
}

// Extended Twisted Edwards Dedicated Doubling Formula in 4M 4S 5a
void EEdDouble(
   const EdPoint *p1,
   EdPoint *p3)
{
	u32 A[EDWARD_LIMBS], B[EDWARD_LIMBS], C[EDWARD_LIMBS], D[EDWARD_LIMBS];
	u32 E[EDWARD_LIMBS], F[EDWARD_LIMBS], G[EDWARD_LIMBS], H[EDWARD_LIMBS];
	u32 p[EDWARD_LIMBS*2];

	// A = X1^2
	Square(EDWARD_LIMBS, p, p1->X);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, A);

	// B = Y1^2
	Square(EDWARD_LIMBS, p, p1->Y);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, B);

	// C = 2 * Z1^2
	Square(EDWARD_LIMBS, p, p1->Z);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, C);
	if (ShiftLeft(EDWARD_LIMBS, C, C, 1))
		Add32(C, EDWARD_LIMBS, EDWARD_C);

	// D = -A
	Negate(EDWARD_LIMBS, D, A);
	Subtract32(D, EDWARD_LIMBS, EDWARD_C);

	// G = D + B, F = G - C, H = D - B = -A - B
	if (Add(G, D, EDWARD_LIMBS, B, EDWARD_LIMBS))
		Add32(G, EDWARD_LIMBS, EDWARD_C);
	if (Subtract(F, G, EDWARD_LIMBS, C, EDWARD_LIMBS))
		Subtract32(F, EDWARD_LIMBS, EDWARD_C);
	if (Subtract(H, D, EDWARD_LIMBS, B, EDWARD_LIMBS))
		Subtract32(H, EDWARD_LIMBS, EDWARD_C);

	// E = (X1 + Y1)^2 - A - B = (X1 + Y1)^2 + H
	if (Add(E, p1->X, EDWARD_LIMBS, p1->Y, EDWARD_LIMBS))
		Add32(E, EDWARD_LIMBS, EDWARD_C);
	Square(EDWARD_LIMBS, p, E);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, E);
	if (Add(E, EDWARD_LIMBS, H, EDWARD_LIMBS))
		Add32(E, EDWARD_LIMBS, EDWARD_C);

	// X3 = E * F
	Multiply(EDWARD_LIMBS, p, E, F);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->X);

	// Y3 = G * H
	Multiply(EDWARD_LIMBS, p, G, H);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->Y);

	// T3 = E * H
	Multiply(EDWARD_LIMBS, p, E, H);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->T);

	// Z3 = F * G
	Multiply(EDWARD_LIMBS, p, F, G);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->Z);
}

// Extended Twisted Edwards Dedicated Doubling Formula in 3M 4S 5a
void EEdDoubleNoT(
   const EdPoint *p1,
   EdPoint *p3)
{
	u32 A[EDWARD_LIMBS], B[EDWARD_LIMBS], C[EDWARD_LIMBS], D[EDWARD_LIMBS];
	u32 E[EDWARD_LIMBS], F[EDWARD_LIMBS], G[EDWARD_LIMBS], H[EDWARD_LIMBS];
	u32 p[EDWARD_LIMBS*2];

	// A = X1^2
	Square(EDWARD_LIMBS, p, p1->X);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, A);

	// B = Y1^2
	Square(EDWARD_LIMBS, p, p1->Y);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, B);

	// C = 2 * Z1^2
	Square(EDWARD_LIMBS, p, p1->Z);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, C);
	if (ShiftLeft(EDWARD_LIMBS, C, C, 1))
		Add32(C, EDWARD_LIMBS, EDWARD_C);

	// D = -A
	Negate(EDWARD_LIMBS, D, A);
	Subtract32(D, EDWARD_LIMBS, EDWARD_C);

	// G = D + B, F = G - C, H = D - B = -A - B
	if (Add(G, D, EDWARD_LIMBS, B, EDWARD_LIMBS))
		Add32(G, EDWARD_LIMBS, EDWARD_C);
	if (Subtract(F, G, EDWARD_LIMBS, C, EDWARD_LIMBS))
		Subtract32(F, EDWARD_LIMBS, EDWARD_C);
	if (Subtract(H, D, EDWARD_LIMBS, B, EDWARD_LIMBS))
		Subtract32(H, EDWARD_LIMBS, EDWARD_C);

	// E = (X1 + Y1)^2 - A - B = (X1 + Y1)^2 + H
	if (Add(E, p1->X, EDWARD_LIMBS, p1->Y, EDWARD_LIMBS))
		Add32(E, EDWARD_LIMBS, EDWARD_C);
	Square(EDWARD_LIMBS, p, E);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, E);
	if (Add(E, EDWARD_LIMBS, H, EDWARD_LIMBS))
		Add32(E, EDWARD_LIMBS, EDWARD_C);

	// X3 = E * F
	Multiply(EDWARD_LIMBS, p, E, F);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->X);

	// Y3 = G * H
	Multiply(EDWARD_LIMBS, p, G, H);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->Y);
/*
	// T3 = E * H
	Multiply(EDWARD_LIMBS, p, E, H);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->T);
*/
	// Z3 = F * G
	Multiply(EDWARD_LIMBS, p, F, G);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, p3->Z);
}

PACKED struct {
	u8 add_index; // nth odd number to add: 0=0,1=1,2=3,3=5,4=7,...
	u8 doubles_after; // number of doubles to perform after add
} MOF_LUT[128] = {
{0,0},{0,1},{1,0},{0,2},{2,0},{1,1},{3,0},{0,3},
{4,0},{2,1},{5,0},{1,2},{6,0},{3,1},{7,0},{0,4},
{8,0},{4,1},{9,0},{2,2},{10,0},{5,1},{11,0},{1,3},
{12,0},{6,1},{13,0},{3,2},{14,0},{7,1},{15,0},{0,5},
{16,0},{8,1},{17,0},{4,2},{18,0},{9,1},{19,0},{2,3},
{20,0},{10,1},{21,0},{5,2},{22,0},{11,1},{23,0},{1,4},
{24,0},{12,1},{25,0},{6,2},{26,0},{13,1},{27,0},{3,3},
{28,0},{14,1},{29,0},{7,2},{30,0},{15,1},{31,0},{0,6},
{32,0},{16,1},{33,0},{8,2},{34,0},{17,1},{35,0},{4,3},
{36,0},{18,1},{37,0},{9,2},{38,0},{19,1},{39,0},{2,4},
{40,0},{20,1},{41,0},{10,2},{42,0},{21,1},{43,0},{5,3},
{44,0},{22,1},{45,0},{11,2},{46,0},{23,1},{47,0},{1,5},
{48,0},{24,1},{49,0},{12,2},{50,0},{25,1},{51,0},{6,3},
{52,0},{26,1},{53,0},{13,2},{54,0},{27,1},{55,0},{3,4},
{56,0},{28,1},{57,0},{14,2},{58,0},{29,1},{59,0},{7,3},
{60,0},{30,1},{61,0},{15,2},{62,0},{31,1},{63,0},{0,7}
};

// Extended Edwards Scalar Multiplication: product = k*P
void EEdScalarMultiply(
	const EdPoint *P,	// Base point
	const u32 *k,		// Scalar multiplicand
	EdPoint *product)	// Result
{
	const int w = 6;

	// Precompute the table
	int precomp_size = 1 << (w - 2);
	EdPoint *precomp = (EdPoint*)alloca(2 * precomp_size * sizeof(EdPoint));

	// Precompute P and -P
	memcpy(precomp, P, sizeof(EdPoint));
	EEdNegate(precomp, precomp+precomp_size);

	// Precompute 2b
	EdPoint doubled;
	EEdDouble(P, &doubled);

	// Precompute +/- odd multiples of b by iteratively adding 2b
	EdPoint *last = precomp;
	int ctr = precomp_size;
	while (--ctr)
	{
		EEdAdd(last, &doubled, last+1);
		++last;
		EEdNegate(last, last+precomp_size);
	}

	// Begin multiplication loop
	bool seen_high_bit = false;
	int limb = EDWARD_LIMBS - 1;
	u32 bits, bits_mask = ((1 << (w + 1)) - 1);
	u32 last_limb = k[limb--];
	int offset = w, doubles_before = 0, doubles_skip = 0;

	for (;;)
	{
		// If still processing bits from current limb of k,
		if (offset <= 32)
		{
			// Select next bits from current limb of k
			bits = last_limb >> (32 - offset);
		}
		else if (limb >= 0)
		{
			// Next bits straddle the previous and next limbs of k
			u32 new_limb = k[limb--];
			offset -= 32;
			bits = (last_limb << offset) | (new_limb >> (32 - offset));
			last_limb = new_limb;
		}
		else if (offset < w + 32)
		{
			// Pad zeroes on the right
			bits = last_limb << (offset - 32);

			// Skip padding - 1 doubles after leaving this loop
			doubles_skip = offset - 32 - 1;
		}
		else break;

		// Mask out garbage bits, leaving only the w+1 bit window behind
		bits &= bits_mask;

		// Select a table entry from the bits
		u32 x = (-(s32)((bits & (1 << w)) >> w) ^ bits) & ((1 << w) - 1);

		if (!x)
			doubles_before += w;
		else
		{
			// Extract the operation for this table entry
			x = (x - 1) >> 1;
			EdPoint *precomputed_point = precomp + MOF_LUT[x].add_index + ((bits & (1 << w)) >> 2);
			int doubles_after = MOF_LUT[x].doubles_after;

			// If we have seen the high bit yet,
			if (seen_high_bit)
			{
				// Perform doubles before addition
				doubles_before += w - doubles_after;

				if (doubles_before)
				{
					for (--doubles_before; doubles_before; --doubles_before)
						EEdDoubleNoT(product, product);
					EEdDouble(product, product);
				}

				// Perform addition or subtraction from the precomputed table
				EEdAddNoT(precomputed_point, product, product);
			}
			else
			{
				// On the first seen bit, product = precomputed point, skip leading doubles
				memcpy(product, precomputed_point, sizeof(EdPoint));

				seen_high_bit = true;
			}

			// Accumulate doubles after addition
			doubles_before = doubles_after;
		}

		// set up offset for next time around
		offset += w;
	}

	// Skip some doubles at the end due to window underrun
	if (doubles_before > doubles_skip)
	{
		doubles_before -= doubles_skip;

		// Perform trailing doubles
		while (doubles_before--);
			EEdDoubleNoT(product, product);
	}
}


//// Server-specific stuff

// Zeroes the private key
TwistedEdwardServer::~TwistedEdwardServer()
{
	OBJCLR(PrivateKey);
}

// Pregenerate the server's private and public keys.  These are shared with clients
// either by being built into the client code or securely delivered via a trusted
// server at runtime.
void TwistedEdwardServer::GenerateOfflineStuff(EdPrivateKey *ServerPrivateKey, EdServerPublicKey *ServerPublicKey)
{
	EdPoint BasePoint;

	// Generate affine (x,y) point on the curve
	do {
		// x = random number between 0 and p-1
		Random::ref()->generate(BasePoint.X, sizeof(BasePoint.X));
		SpecialModulus(BasePoint.X, EDWARD_LIMBS, EDWARD_C, EDWARD_LIMBS, BasePoint.X);

		// y = solution of curve from x (very slow!)
		EEdSolveY(BasePoint.X, BasePoint.Y);
	} while (!EEdIsValidXY(BasePoint.X, BasePoint.Y));

	// Z = 1
	Set32(BasePoint.Z, EDWARD_LIMBS, 1);

	// P = 4P
	EEdDoubleNoT(&BasePoint, &BasePoint); // T unused for this calculation
	EEdDouble(&BasePoint, &BasePoint); // T unused for this one also, but it generates T

	// Generate random private key
	Random::ref()->generate(ServerPrivateKey->Bytes, sizeof(ServerPrivateKey->Bytes));

	// Compute k*P (slow!)
	EdPoint PublicProduct;
	EEdScalarMultiply(&BasePoint, ServerPrivateKey->Limbs, &PublicProduct);

	// Convert product and base to affine coordinates for 50% compression
	EEdGetAffineXY(&PublicProduct, ServerPublicKey->Product_x.Limbs, ServerPublicKey->Product_y.Limbs);
	EEdGetAffineXY(&BasePoint, ServerPublicKey->BasePoint_x.Limbs, ServerPublicKey->BasePoint_y.Limbs);

	// Convert it all to endian-neutral byte arrays
	ToLittleEndian(ServerPublicKey->BasePoint_x.Limbs, EDWARD_LIMBS);
	ToLittleEndian(ServerPublicKey->BasePoint_y.Limbs, EDWARD_LIMBS);
	ToLittleEndian(ServerPrivateKey->Limbs, EDWARD_LIMBS);
	ToLittleEndian(ServerPublicKey->Product_x.Limbs, EDWARD_LIMBS);
	ToLittleEndian(ServerPublicKey->Product_y.Limbs, EDWARD_LIMBS);
}

// Called on start-up to initialize the object
void TwistedEdwardServer::SetPrivateKey(const EdPrivateKey *ServerPrivateKey)
{
	Set(PrivateKey, EDWARD_LIMBS, ServerPrivateKey->Limbs);

	// Fix byte order
	FromLittleEndian(PrivateKey, EDWARD_LIMBS);
}

// Compute a shared secret from a client's public key.  Thread-safe.  False on invalid input.
bool TwistedEdwardServer::ComputeSharedSecret(const EdClientPublicKey *ClientPublicKey, EdSharedSecret *SharedSecret)
{
	// Decompress client's public point
	EdPoint ClientPublicPoint;

	Set(ClientPublicPoint.X, EDWARD_LIMBS, ClientPublicKey->Product_x.Limbs);
	Set(ClientPublicPoint.Y, EDWARD_LIMBS, ClientPublicKey->Product_y.Limbs);

	// Fix byte order
	FromLittleEndian(ClientPublicPoint.X, EDWARD_LIMBS);
	FromLittleEndian(ClientPublicPoint.Y, EDWARD_LIMBS);

	// Verify the point is on the curve (relatively fast)
	if (!EEdIsValidXY(ClientPublicPoint.X, ClientPublicPoint.Y))
		return false;

	// T = XY
	u32 p[EDWARD_LIMBS*2];
	Multiply(EDWARD_LIMBS, p, ClientPublicPoint.X, ClientPublicPoint.Y);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, ClientPublicPoint.T);

	// Z = 1
	Set32(ClientPublicPoint.Z, EDWARD_LIMBS, 1);

	// Compute k*P (slow!)
	EdPoint SharedProduct;
	EEdScalarMultiply(&ClientPublicPoint, PrivateKey, &SharedProduct);

	// Get the affine x coordinate of the product; should be the same one computed by the client
	u32 SharedSecretRaw[EDWARD_LIMBS];
	EEdGetAffineX(&SharedProduct, SharedSecretRaw);

	// Convert it into an endian-neutral byte array
	ToLittleEndian(SharedSecretRaw, EDWARD_LIMBS);

	// Hash the byte array
	DigestSHA256 sha256;
	sha256.performDigest(SharedSecretRaw, sizeof(SharedSecretRaw));

	// Return the hash, which is an endian-neutral byte array
	memcpy(SharedSecret->Bytes, sha256.endDigest(), sizeof(SharedSecret->Bytes));
	return true;
}


//// Client-specific stuff

// Called on start-up to initialize the object
void TwistedEdwardClient::SetServerPublicKey(const EdServerPublicKey *ServerPublicKey)
{
	// Decompress server's public point and the base point
	Set(BasePoint.X, EDWARD_LIMBS, ServerPublicKey->BasePoint_x.Limbs);
	Set(BasePoint.Y, EDWARD_LIMBS, ServerPublicKey->BasePoint_y.Limbs);
	Set(ServerPublicPoint.X, EDWARD_LIMBS, ServerPublicKey->Product_x.Limbs);
	Set(ServerPublicPoint.Y, EDWARD_LIMBS, ServerPublicKey->Product_y.Limbs);

	// Fix byte order
	FromLittleEndian(BasePoint.X, EDWARD_LIMBS);
	FromLittleEndian(BasePoint.Y, EDWARD_LIMBS);
	FromLittleEndian(ServerPublicPoint.X, EDWARD_LIMBS);
	FromLittleEndian(ServerPublicPoint.Y, EDWARD_LIMBS);

	// T = XY
	u32 p[EDWARD_LIMBS*2];
	Multiply(EDWARD_LIMBS, p, BasePoint.X, BasePoint.Y);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, BasePoint.T);
	Multiply(EDWARD_LIMBS, p, ServerPublicPoint.X, ServerPublicPoint.Y);
	SpecialModulus(p, EDWARD_LIMBS*2, EDWARD_C, EDWARD_LIMBS, ServerPublicPoint.T);

	// Z = 1
	Set32(BasePoint.Z, EDWARD_LIMBS, 1);
	Set32(ServerPublicPoint.Z, EDWARD_LIMBS, 1);
}

// Compute a shared secret and client's public key
void TwistedEdwardClient::ComputeSharedSecret(EdClientPublicKey *ClientPublicKey, EdSharedSecret *SharedSecret)
{
	// Generate a private key
	u32 PrivateKey[EDWARD_LIMBS];
	Random::ref()->generate(PrivateKey, sizeof(PrivateKey));

	// Compute "private key * base point = client public key" (slow!)
	EdPoint ClientProduct;
	EEdScalarMultiply(&BasePoint, PrivateKey, &ClientProduct);

	// Get the affine (x,y) product point
	EEdGetAffineXY(&ClientProduct, ClientPublicKey->Product_x.Limbs, ClientPublicKey->Product_y.Limbs);

	// Convert it into an endian-neutral byte array
	ToLittleEndian(ClientPublicKey->Product_x.Limbs, EDWARD_LIMBS);
	ToLittleEndian(ClientPublicKey->Product_y.Limbs, EDWARD_LIMBS);

	// Compute "private key * server public = shared secret point" (slow!)
	EdPoint SharedProduct;
	EEdScalarMultiply(&ServerPublicPoint, PrivateKey, &SharedProduct);

	// Get the affine x coordinate of the product; should be the same one computed by the server
	u32 SharedSecretRaw[EDWARD_LIMBS];
	EEdGetAffineX(&SharedProduct, SharedSecretRaw);

	// Convert it into an endian-neutral byte array
	ToLittleEndian(SharedSecretRaw, EDWARD_LIMBS);

	// Hash the byte array
	DigestSHA256 sha256;
	sha256.performDigest(SharedSecretRaw, sizeof(SharedSecretRaw));

	// Return the hash, which is an endian-neutral byte array
	memcpy(SharedSecret->Bytes, sha256.endDigest(), sizeof(SharedSecret->Bytes));
}
