#ifndef ECC_HPP
#define ECC_HPP

#include "Platform.hpp"

/*
	Elliptic Curve Cryptography for Diffie-Hellman key exchange

	Example usage in UDP:

		 Lines marked with (*) are provided by this module.

		Offline:

			(*) Programmer pre-generates Server Public Key (128 by)
			    and embeds it into the client and server executables.

		Online:

			Client wants to connect.

			Client listens for ICMP failure messages to determine if the server
			is running or not at the remote address.  Much faster than a timeout.

			[UDP/IP] Client->Server transmit Hello (~8 by).

			(*) Client generates Client Public Key (64 by) and Shared Secret (32 by).

			Server validates Hello.

			[UDP/IP] Server->Client transmit Cookie (8 by).
			This would be stateless so no connection has been made yet.

			Client stops accepting ICMP failure messages.

			[UDP/IP] Client->Server transmit Client Public Key (64 by) and Cookie (8 by).

			Server validates Cookie.
			(*) Server validates the Client Public Key.

			(*) Server computes Shared Secret (takes about 1 ms on server hardware).

			[UDP/IP] Server->Client transmits Session Key encrypted with Shared Secret.
			Example: Nonce | Encrypted { Hash(Session Key|...) | Session Key | ... }
					 This would be about 40 bytes minimum.

			Server marks the client as being connected and remembers the client's
			public key and cookie and caches its ~40 byte response so if the client
			lost the server's response the server can quickly retransmit it.

			Client decrypts and verifies the hash of the Session Key.

			Now both sides have the Session Key and the server has accepted the client.

	Example network architecture:

		I would recommend building a network architecture that has:
			1 List server
			World servers
			Update servers
			1 Database server

		The client would contact the List server on start-up to get a list of all
		the World/Update servers and determine if it needs any updates.

		In this case, the client will only need to statically know the key of the List
		server.  Then, the List server can tell the Client the Server Public Key for each
		of the World/Update servers.

	Example code:

		TwistedEdwardServer server;
		TwistedEdwardClient client;

		EdPrivateKey server_private_key;
		EdServerPublicKey server_public_key;
		EdClientPublicKey client_public_key;
		EdSharedSecret client_shared_secret;
		EdSharedSecret server_shared_secret;

	--done by the programmer once--

		// OFFLINE PRECOMPUTATION: Generating server private and public keys
		server.GenerateOfflineStuff(&server_private_key, &server_public_key);

	--client is compiled with the server's public key--

	--client and server programs are executed--

		// ON SERVER START-UP: Loading server private key
		server.SetPrivateKey(&server_private_key);

		// ON CLIENT START-UP: Loading server public key
		client.SetServerPublicKey(&server_public_key);

	--client is attempting to connect to server--

		// CLIENT: ONLINE NOT TIME-CRITICAL: Computing shared secret
		client.ComputeSharedSecret(&client_public_key, &client_shared_secret);

	--server receives client_public_key here over the internet--

		// SERVER: ONLINE AND TIME-CRITICAL: Computing shared secret
		server.ComputeSharedSecret(&client_public_key, &server_shared_secret);

	--server should now send a session key encrypted with the Shared Secret--

		Hurray we have a secure session! 8D
*/


//// Curve constants

// After changing the number of limbs, C and D must be recalculated.
// After changing C, D must be recalculated.
#define EDWARD_LIMBS 8 /* 256 bits */
#define EDWARD_C 189 /* field prime modulus (2^N - c) parameter c */
#define EDWARD_D 321 /* curve (yy-xx=1+dxxyy) parameter d */

/*
	Other values of d that work for c = 189, sorted by number of points (NP):

	NP/4=28948022309329048855892746252171976963449087812878447882682178384830544116457 for d=321
	NP/4=28948022309329048855892746252171976963364137860193714257833303761363762534599 for d=3567
	NP/4=28948022309329048855892746252171976963356671321081301322781900858624717722023 for d=2087
	NP/4=28948022309329048855892746252171976963209239424227502814070775171816900352269 for d=4238
	NP/4=28948022309329048855892746252171976963194848828905762674632598658951271884767 for d=889
	NP/4=28948022309329048855892746252171976963193685037190250905176734116833185051207 for d=2715
	NP/4=28948022309329048855892746252171976963171131370530754794550345168351098815157 for d=3272
	NOTE: These are all prime and very close to p/4.
*/


//// Data structures

struct EdPoint {
	u32 X[EDWARD_LIMBS];
	u32 Y[EDWARD_LIMBS];
	u32 T[EDWARD_LIMBS];
	u32 Z[EDWARD_LIMBS];
};

// NOTE: All of these packed structures are endian-neutral byte arrays

PACKED union EdInt
{
	u32 Limbs[EDWARD_LIMBS];
	u8 Bytes[EDWARD_LIMBS * sizeof(u32)];
};

typedef EdInt EdPrivateKey;

PACKED struct EdSharedSecret
{
	u8 Bytes[32];
};

PACKED struct EdServerPublicKey
{
	// Affine coordinates
	EdInt BasePoint_x, BasePoint_y;
	EdInt Product_x, Product_y;
};

PACKED struct EdClientPublicKey
{
	// Affine coordinates
	EdInt Product_x, Product_y;
};

/*
	class TwistedEdwardServer

	Generates server public and private keys.  Computes the shared secret between
	the server and the client based on the client's public key.

	One single instance of this object is sufficient to service all the clients.
	The online members are thread-safe.
*/
class TwistedEdwardServer
{
	u32 PrivateKey[EDWARD_LIMBS];

public:
	// Zeroes the private key
	~TwistedEdwardServer();

	// Pregenerate the server's private and public keys.  These are shared with clients
	// either by being built into the client code or securely delivered via a trusted
	// server at runtime.
	static void GenerateOfflineStuff(EdPrivateKey *ServerPrivateKey, EdServerPublicKey *ServerPublicKey);

	// Called on start-up to initialize the object
	void SetPrivateKey(const EdPrivateKey *ServerPrivateKey);

	// Compute a shared secret from a client's public key.  Thread-safe.  False on invalid input.
	bool ComputeSharedSecret(const EdClientPublicKey *ClientPublicKey, EdSharedSecret *SharedSecret);
};

/*
	class TwistedEdwardClient
*/
class TwistedEdwardClient
{
	EdPoint BasePoint, ServerPublicPoint;

public:
	// Called on start-up to initialize the object
	void SetServerPublicKey(const EdServerPublicKey *ServerPublicKey);

	// Compute a shared secret and client's public key
	void ComputeSharedSecret(EdClientPublicKey *ClientPublicKey, EdSharedSecret *SharedSecret);
};


//// Math

// Returns x = X/Z
void EEdGetAffineX(
	const EdPoint *p,
	u32 *x);

// Returns (x,y) = (X/Z,Y/Z)
void EEdGetAffineXY(
	const EdPoint *p1,
	u32 *x,
	u32 *y);

// Solve for Y given the X point on a curve
void EEdSolveY(
	const u32 *x,
	u32 *y);

// Verify that the point (x,y) exists on the given curve
bool EEdIsValidXY(
	const u32 *x,
	const u32 *y);

// Extended Twisted Edwards Negation Formula in 2a
void EEdNegate(
	const EdPoint *p1,
	EdPoint *p3);

// Extended Twisted Edwards Unified Addition Formula (works when both inputs are the same) in 8M 1D 8A
void EEdAdd(
	const EdPoint *p1,
	const EdPoint *p2,
	EdPoint *p3);

// Extended Twisted Edwards Unified Addition Formula (works when both inputs are the same) in 7M 1D 8A
void EEdAddNoT(
	const EdPoint *p1,
	const EdPoint *p2,
	EdPoint *p3);

// Extended Twisted Edwards Dedicated Doubling Formula in 4M 4S 5a
void EEdDouble(
	const EdPoint *p1,
	EdPoint *p3);

// Extended Twisted Edwards Dedicated Doubling Formula in 3M 4S 5a
void EEdDoubleNoT(
	const EdPoint *p1,
	EdPoint *p3);

// Extended Edwards Scalar Multiplication: product = k*P
void EEdScalarMultiply(
	const EdPoint *P,	// Base point
	const u32 *k,		// Scalar multiplicand
	EdPoint *product);	// Result

#endif // ECC_HPP
