#include "Logging.hpp"

#include "Settings.hpp"
#include "Clock.hpp"

#include <ctime>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <process.h>
using namespace std;


//#define BREAK_ON_ERROR /* Generates breakpoint interrupt before throwing exceptions */


static const char *EVENT_NAME[5] = { "Inane", "Info", "Warn", "Oops", "Fatal" };
//static const char *SHORT_EVENT_NAME[5] = { ".", "I", "W", "!", "F" };


//// Free functions

region_string HexDumpString(const void *vdata, u32 bytes)
{
	/* xxxx  xx xx xx xx xx xx xx xx  xx xx xx xx xx xx xx xx   aaaaaaaaaaaaaaaa*/

	const u8 *data = (const u8*)vdata;
	u32 ii, offset;

	char ascii[17];
	ascii[16] = 0;

	region_ostringstream oss;

	for (offset = 0; offset < bytes; offset += 16)
	{
		oss << endl << setfill('0') << hex << setw(4) << offset << "  ";

		for (ii = 0; ii < 16; ++ii)
		{
			if (ii == 8)
				oss << ' ';

			if (offset + ii < bytes)
			{
				u8 ch = data[offset + ii];

				oss << setw(2) << (u32)ch << ' ';
				ascii[ii] = (ch >= ' ' && ch <= '~') ? ch : '.';
			}
			else
			{
				oss << "   ";
				ascii[ii] = 0;
			}
		}

		oss << " " << ascii;
	}

	return oss.str();
}

void OutputConsoleDebug(LogEvent *logEvent)
{
	region_ostringstream oss;
	oss << "[" << Clock::format("%b %d %H:%M") << "] <" << logEvent->subsystem << "> "
		<< logEvent->msg.str().c_str() << endl;
	cout << oss.str();
	OutputDebugStringA(oss.str().c_str());
}


//// LogEvent

LogEvent::LogEvent()
{
	subsystem = 0;
	severity = LVL_FATAL;
}

LogEvent::LogEvent(const char *subsystem, EventSeverity severity)
{
	this->subsystem = subsystem;
	this->severity = severity;
}


//// Logging

unsigned int WINAPI EventProcessorThread(void *param)
{
	( (Logging*)param )->EventDequeueProcessor();
	return 0;
}

Logging::Logging()
{
	callback = 0;
	log_threshold = LVL_INANE;

	hThread = (HANDLE)_beginthreadex(0, 0, EventProcessorThread, this, 0, 0);
}

__declspec(dllexport) void SetLogCallback(LogCallback cb)
{
	Logging::ref()->SetLogCallback(cb);
}

void Logging::Initialize()
{
	log_threshold = Settings::ii->getInt("Log.Threshold", LVL_INANE);
}

void Logging::Shutdown()
{
	if (hThread != (HANDLE)-1)
	{
		queue.Enqueue(new (RegionAllocator::ii) LogEvent);

		WaitForSingleObject(hThread, INFINITE);

		CloseHandle(hThread);

		// Dequeue and process late messages
		LogEvent *logEvent;
		while ((logEvent = queue.Dequeue()) && logEvent->subsystem)
			OutputConsoleDebug(logEvent);
	}
}

void Logging::QueueEvent(LogEvent *logEvent)
{
	queue.Enqueue(logEvent);
}

void Logging::EventDequeueProcessor()
{
	LogEvent *logEvent;

	while ((logEvent = queue.DequeueWait()) && logEvent->subsystem)
	{
		if (callback)
			callback(EVENT_NAME[logEvent->severity], logEvent->subsystem, logEvent->msg.str().c_str());
		else
			OutputConsoleDebug(logEvent);
	}
}


//// Recorder

Recorder::Recorder(const char *subsystem, EventSeverity severity)
{
	logEvent = new (RegionAllocator::ii) LogEvent(subsystem, severity);

	if (!logEvent)
		throw std::runtime_error("Out of region memory while recording an event");
}

Recorder::~Recorder()
{
	Logging::ii->QueueEvent(logEvent);
}


//// Enforcer

Enforcer::Enforcer(const char *locus)
{
	oss << locus;
}

Enforcer::~Enforcer()
{
#if defined(BREAK_ON_ERROR)
# if defined(__GNUC__)
	__asm__ ( "int $3" );
# elif defined(_MSC_VER)
	__asm int 3;
# endif
#endif

	throw std::runtime_error(oss.str());
}
